/*
 * Copyright 2014-2016 Simon Edwards <simon@simonzone.com>
 *
 * This source code is licensed under the MIT license which is detailed in the LICENSE.txt file.
 */

import { BulkFileMetadata, EnvironmentMap } from 'extraterm-extension-api';
import * as Electron from 'electron';
const ipc = Electron.ipcRenderer;

import {BulkFileIdentifier} from '../main_process/bulk_file_handling/BulkFileStorage';
import * as Messages from '../WindowMessages';
import * as config from '../Config';
import {Logger, getLogger} from "extraterm-logging";
import { ExtensionMetadata } from '../ExtensionMetadata';
import { ThemeType } from '../theme/Theme';
import { KeybindingsFile } from '../keybindings/KeybindingsFile';

const _log = getLogger("WebIPC");

const DEBUG = false;
const DEBUG_INCOMING = false;
const DEBUG_OUTGOING = false;

/**
 * Start IPC.
 */
export function start(): void {
  ipc.on(Messages.CHANNEL_NAME, handleAsyncIpc);
}

type ResolveFunc = (msg: Messages.Message) => void;

// This queue is used to route incoming IPC messages to expectant promise objects.
let promiseQueue: {promiseResolve: ResolveFunc, messageType: Messages.MessageType}[] = [];

export type Handler = (msg: Messages.Message) => void;
const defaultHandlers: { messageType: Messages.MessageType, handler: Handler }[] = [];

/**
 * Register a default handler for a message type
 * 
 * @param type    the message type this handler should be used for.
 * @param handler the handler itself.
 */
export function registerDefaultHandler(type: Messages.MessageType, handler: Handler): void {
  defaultHandlers.push( { messageType: type, handler: handler} );
}

function handleAsyncIpc(senderInfo: any, detail: any): void {
  const msg: Messages.Message = detail;
  if (DEBUG_INCOMING) {
    _log.debug("incoming: ", msg);
  }

  const matchingPromises = promiseQueue.filter( p => p.messageType === msg.type );
  const nonMatchingPromises = promiseQueue.filter( p => p.messageType !== msg.type );
  promiseQueue = nonMatchingPromises;
  
  matchingPromises.forEach( tup => {
    tup.promiseResolve(msg);
  });
  
  if (matchingPromises.length === 0) {
    // Fall back on the default handlers.
    defaultHandlers.filter( tup => tup.messageType === msg.type ).forEach( tup => {
      tup.handler(msg);
    });
  }
}

function request(msg: Messages.Message, replyType: Messages.MessageType): Promise<Messages.Message> {
  if (DEBUG_OUTGOING) {
    _log.debug("request: ", msg);
  }
  ipc.send(Messages.CHANNEL_NAME, msg);
  return new Promise<Messages.Message>( (resolve,cancel) => {
    promiseQueue.push( { promiseResolve: resolve, messageType: replyType } );
  });
}

export function requestConfig(key: config.ConfigKey): Promise<Messages.ConfigMessage> {
  const msg: Messages.ConfigRequestMessage = {type: Messages.MessageType.CONFIG_REQUEST, key};
  return <Promise<Messages.ConfigMessage>> request(msg, Messages.MessageType.CONFIG);
}

export function requestThemeList(): Promise<Messages.ThemeListMessage> {
  const msg: Messages.ThemeListRequestMessage = {type: Messages.MessageType.THEME_LIST_REQUEST};
  return <Promise<Messages.ThemeListMessage>> request(msg, Messages.MessageType.THEME_LIST);
}

export function requestThemeContents(themeType: ThemeType): Promise<Messages.ThemeContentsMessage> {
  if (DEBUG) {
    _log.debug("requestThemeContents(): ", themeType);
  }
  const msg: Messages.ThemeContentsRequestMessage = {
    type: Messages.MessageType.THEME_CONTENTS_REQUEST,
    themeType
  };
  return <Promise<Messages.ThemeContentsMessage>> request(msg, Messages.MessageType.THEME_CONTENTS);
}

export function rescanThemes(): void {
  const msg: Messages.ThemeRescan = {type: Messages.MessageType.THEME_RESCAN };
  ipc.send(Messages.CHANNEL_NAME, msg);
}

export function requestPtyCreate(sessionUuid: string, extraEnv: EnvironmentMap, columns: number, rows: number
    ): Promise<Messages.CreatedPtyMessage> {
      
  const msg: Messages.CreatePtyRequestMessage = {
    type: Messages.MessageType.PTY_CREATE,
    sessionUuid,
    columns: columns,
    rows: rows,
    env: extraEnv
  };
  return <Promise<Messages.CreatedPtyMessage>> request(msg, Messages.MessageType.PTY_CREATED);
}

export function ptyInput(id: number, data: string): void {
  const msg: Messages.PtyInput = {type: Messages.MessageType.PTY_INPUT, id: id, data: data };
  ipc.send(Messages.CHANNEL_NAME, msg);
}

export function ptyOutputBufferSize(id: number, size: number): void {
  const msg: Messages.PtyOutputBufferSize = {type: Messages.MessageType.PTY_OUTPUT_BUFFER_SIZE, id, size };
  ipc.send(Messages.CHANNEL_NAME, msg);
}

export function ptyResize(id: number, columns: number, rows: number): void {
  const msg: Messages.PtyResize = {type: Messages.MessageType.PTY_RESIZE, id: id, columns: columns, rows: rows };
  ipc.send(Messages.CHANNEL_NAME, msg);
}

export function ptyClose(id: number): void {
  const msg: Messages.PtyCloseRequest = {type: Messages.MessageType.PTY_CLOSE_REQUEST, id: id };
  ipc.send(Messages.CHANNEL_NAME, msg);
}

export function devToolsRequest(open: boolean): void {
  const msg: Messages.DevToolsRequestMessage = { type: Messages.MessageType.DEV_TOOLS_REQUEST, open: open };
  ipc.send(Messages.CHANNEL_NAME, msg);
}

export function clipboardWrite(text: string): void {
  const msg: Messages.ClipboardWriteMessage = { type: Messages.MessageType.CLIPBOARD_WRITE, text: text };
  ipc.send(Messages.CHANNEL_NAME, msg);    
}

export function clipboardReadRequest(): void {
  const msg: Messages.ClipboardReadRequestMessage = { type: Messages.MessageType.CLIPBOARD_READ_REQUEST };
  ipc.send(Messages.CHANNEL_NAME, msg);      
}

export function windowCloseRequest(): void {
  const msg: Messages.WindowCloseRequestMessage = { type: Messages.MessageType.WINDOW_CLOSE_REQUEST };
  ipc.send(Messages.CHANNEL_NAME, msg);  
}

export function sendConfig(key: config.ConfigKey, config: any): void {
  const msg: Messages.ConfigMessage = { type: Messages.MessageType.CONFIG, key, config };
  ipc.send(Messages.CHANNEL_NAME, msg);  
}

export function requestNewTag(): Promise<Messages.NewTagMessage> {
  const msg: Messages.NewTagRequestMessage = {type: Messages.MessageType.NEW_TAG_REQUEST, async: true};
  return <Promise<Messages.NewTagMessage>> request(msg, Messages.MessageType.NEW_TAG);
}

export function requestNewTagSync(): string {
  const msg: Messages.NewTagRequestMessage = {type: Messages.MessageType.NEW_TAG_REQUEST, async: false};
  const event = <any> ipc.sendSync(Messages.CHANNEL_NAME, msg);
  const newTagMessage = <Messages.NewTagMessage> event;
  return newTagMessage.tag;
}

export function windowMinimizeRequest(): void {
  const msg: Messages.WindowMinimizeRequestMessage = { type: Messages.MessageType.WINDOW_MINIMIZE_REQUEST };
  ipc.send(Messages.CHANNEL_NAME, msg);  
}

export function windowMaximizeRequest(): void {
  const msg: Messages.WindowMaximizeRequestMessage = { type: Messages.MessageType.WINDOW_MAXIMIZE_REQUEST };
  ipc.send(Messages.CHANNEL_NAME, msg);  
}

export function createBulkFileSync(metadata: BulkFileMetadata, size: number): {identifier: BulkFileIdentifier, url: string} {
  const msg: Messages.BulkFileCreateMessage = {type: Messages.MessageType.BULK_FILE_CREATE, metadata, size};
  const event = <any> ipc.sendSync(Messages.CHANNEL_NAME, msg);
  const createdBulkFileMessage = <Messages.BulkFileCreatedResponseMessage> event;
  return {identifier: createdBulkFileMessage.identifier, url: createdBulkFileMessage.url};
}

export function writeBulkFile(identifier: BulkFileIdentifier, data: Buffer): void {
  const msg: Messages.BulkFileWriteMessage = {type: Messages.MessageType.BULK_FILE_WRITE, identifier, data};
  ipc.send(Messages.CHANNEL_NAME, msg);
}

export function closeBulkFile(identifier: BulkFileIdentifier, success: boolean): void {
  const msg: Messages.BulkFileCloseMessage = {type: Messages.MessageType.BULK_FILE_CLOSE, identifier, success};
  ipc.send(Messages.CHANNEL_NAME, msg);
}

export function refBulkFile(identifier: BulkFileIdentifier): void {
  const msg: Messages.BulkFileRefMessage = {type: Messages.MessageType.BULK_FILE_REF, identifier};
  ipc.send(Messages.CHANNEL_NAME, msg);
}

export function derefBulkFile(identifier: BulkFileIdentifier): void {
  const msg: Messages.BulkFileDerefMessage = {type: Messages.MessageType.BULK_FILE_DEREF, identifier};
  ipc.send(Messages.CHANNEL_NAME, msg);
}

export function requestExtensionMetadataSync(): ExtensionMetadata[] {
  const msg: Messages.ExtensionMetadataRequestMessage = {type: Messages.MessageType.EXTENSION_METADATA_REQUEST};
  const event = <any> ipc.sendSync(Messages.CHANNEL_NAME, msg);
  const extensionMetadataMessage = <Messages.ExtensionMetadataMessage> event;
  return extensionMetadataMessage.extensionMetadata;
}

export function keybindingsCopy(sourceName: string, destName: string): void {
  const msg: Messages.KeybindingsCopyMessage = {type: Messages.MessageType.COPY_KEYBINDINGS, sourceName, destName};
  ipc.send(Messages.CHANNEL_NAME, msg);
}

export function keybindingsRequestRead(name: string): Promise<Messages.KeybindingsReadMessage> {
  const msg: Messages.KeybindingsReadRequestMessage = {
    type: Messages.MessageType.READ_KEYBINDINGS_REQUEST,
    name
  };
  return <Promise<Messages.KeybindingsReadMessage>> request(msg, Messages.MessageType.READ_KEYBINDINGS);
}

export function keybindingsRename(sourceName: string, destName: string): void {
  const msg: Messages.KeybindingsRenameMessage = {type: Messages.MessageType.RENAME_KEYBINDINGS, sourceName, destName};
  ipc.send(Messages.CHANNEL_NAME, msg);
}

export function keybindingsDelete(name: string): void {
  const msg: Messages.KeybindingsDeleteMessage = {type: Messages.MessageType.DELETE_KEYBINDINGS, name: name};
  ipc.send(Messages.CHANNEL_NAME, msg);
}

export function keybindingsUpdate(name: string, keybindings: KeybindingsFile): void {
  const msg: Messages.KeybindingsUpdateMessage = {type: Messages.MessageType.UPDATE_KEYBINDINGS, name, keybindings};
  ipc.send(Messages.CHANNEL_NAME, msg);
}

export function enableGlobalKeybindings(enabled: boolean): void {
  const msg: Messages.GlobalKeybindingsEnableMessage = {type: Messages.MessageType.GLOBAL_KEYBINDINGS_ENABLE, enabled};
  ipc.send(Messages.CHANNEL_NAME, msg);
}

export function requestTerminalTheme(id: string): Promise<Messages.TerminalThemeMessage> {
  const msg: Messages.TerminalThemeRequestMessage = {
    type: Messages.MessageType.TERMINAL_THEME_REQUEST,
    id
  };
  return <Promise<Messages.TerminalThemeMessage>> request(msg, Messages.MessageType.TERMINAL_THEME);
}
